/*
Copyright  2004-2007 Eland Systems All Rights Reserved.

   1. Redistribution and use in source and binary forms must retain the above
   copyright notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

   3. The name of Eland Systems may not be used to endorse or promote products
   derived from this software without specific prior written permission.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sysexits.h>
#include <syslog.h>
#include <pthread.h>

#include <netdb.h>
#include <arpa/nameser.h>
#include <resolv.h>

#ifndef LINUX
#include <sys/queue.h>
#else
#include "queue.h"
#endif

#include "lock.h"
#include "util.h"

#ifdef POPAUTH
# include <db.h>
# include <fcntl.h>
#endif

#include <regex.h>

#ifndef	REG_BASIC
#define	REG_BASIC 0
#endif

#define LINELENGTH 1024

#define SCAMCONF "/etc/mail/scam.conf"

static unsigned char lowercase[256] = {
    0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
   16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31,
  ' ','!','"','#','$','%','&', 39,'(',')','*','+',',','-','.','/',
  '0','1','2','3','4','5','6','7','8','9',':',';','<','=','>','?',
  '@','a','b','c','d','e','f','g','h','i','j','k','l','m','n','o',
  'p','q','r','s','t','u','v','w','x','y','z','[', 92,']','^','_',
  '`','a','b','c','d','e','f','g','h','i','j','k','l','m','n','o',
  'p','q','r','s','t','u','v','w','x','y','z','{','|','}','~',127,
  128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143,
  144,145,146,147,148,149,150,151,152,153,154,155,156,157,158,159,
  160,161,162,163,164,165,166,167,168,169,170,171,172,173,174,175,
  176,177,178,179,180,181,182,183,184,185,186,187,188,189,190,191,
  192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,
  208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223,
  224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239,
  240,241,242,243,244,245,246,247,248,249,250,251,252,253,254,255};

#define LPC(x) (lowercase[(unsigned char)(x)])

struct rrmx {
	int pref;
	char *hostname;
};

struct entry {
	char	*c;
	SLIST_ENTRY(entry)	entries;
};

SLIST_HEAD(, entry) gwhitelist;
SLIST_HEAD(, entry) dnsbllist;

pthread_rwlock_t gwhite_lock;

struct CIDR {
	unsigned long ip;
	short int masklen;
	SLIST_ENTRY(CIDR)	cidrs;
};

SLIST_HEAD(, CIDR)	cidrlist;

struct sre {
	regex_t	 re;
	short int cat;
	SLIST_ENTRY(sre) sres;
};

SLIST_HEAD(, sre)	whiterelist;
SLIST_HEAD(, sre)	greyhostrelist;

struct ientry {
	unsigned int	i;
	SLIST_ENTRY(ientry) ientries;
};

SLIST_HEAD(, ientry) greyasnlist;
SLIST_HEAD(, ientry) whiteasnlist;

size_t
strlcpy(char *dst, const char *src, size_t siz)
{
	register char *d = dst;
	register const char *s = src;
	register size_t n = siz;

	/* Copy as many bytes as will fit */
	if (n != 0 && --n != 0) {
		do {
			if ((*d++ = *s++) == 0)
				break;
		} while (--n != 0);
	}

	/* Not enough room in dst, add NUL and traverse rest of src */
	if (n == 0) {
		if (siz != 0)
			*d = '\0';		/* NUL-terminate dst */
		while (*s++)
			;
	}

	return(s - src - 1);	/* count does not include NUL */
}

void gwhite_init()
{
	int ret;

	if ((ret = pthread_rwlock_init(&gwhite_lock, NULL)) != 0)
	{
		syslog(LOG_ERR,  "pthread_rwlock_init failed: %s", strerror(ret));
		exit(EX_OSERR);
	}
	SLIST_INIT(&cidrlist);
}

void whitere_init()
{
	SLIST_INIT(&whiterelist);
}

void greyhostre_init()
{
	SLIST_INIT(&greyhostrelist);
}

int gwhite_readconf()
{
	FILE *fh;
	struct entry *ent;
	char line[LINELENGTH + 1];
	char buf[LINELENGTH + 1];
	struct CIDR *entcidr;

	struct sre *entre;
	struct ientry *ient;
	int a;

	int lline;
	int lineno = 0;

	WRLOCK(gwhite_lock);
	while (!SLIST_EMPTY(&gwhitelist))
	{
		ent = SLIST_FIRST(&gwhitelist);
		free(ent->c);
		SLIST_REMOVE_HEAD(&gwhitelist, entries);
		free(ent);
	}

	while (!SLIST_EMPTY(&cidrlist))
	{
		entcidr = SLIST_FIRST(&cidrlist);
		SLIST_REMOVE_HEAD(&cidrlist, cidrs);
		free(entcidr);
	}

	while (!SLIST_EMPTY(&whiterelist))
	{
		entre = SLIST_FIRST(&whiterelist);
		SLIST_REMOVE_HEAD(&whiterelist, sres);
		free(entre);
	}

	while (!SLIST_EMPTY(&whiteasnlist))
	{
		ient = SLIST_FIRST(&whiteasnlist);
		SLIST_REMOVE_HEAD(&whiteasnlist, ientries);
		free(ient);
	}
	while (!SLIST_EMPTY(&greyasnlist))
	{
		ient = SLIST_FIRST(&greyasnlist);
		SLIST_REMOVE_HEAD(&greyasnlist, ientries);
		free(ient);
	}

	while (!SLIST_EMPTY(&dnsbllist))
	{
		ent = SLIST_FIRST(&dnsbllist);
		free(ent->c);
		SLIST_REMOVE_HEAD(&dnsbllist, entries);
		free(ent);
	}

	if ((fh = fopen( SCAMCONF, "r")) == NULL)
	{
		UNLOCK(gwhite_lock);
		return 1;
	}

	syslog (LOG_INFO, "reading greylist conf");

	while( fgets(line, LINELENGTH, fh) )
	{
		lineno++;
		if (line[0] == '#')
			continue;

		lline = strlen(line);
		if (lline > 16)
		{
			unsigned int ip1 = 0;
			unsigned int ip2 = 0;
			unsigned int ip3 = 0;
			unsigned int ip4 = 0;
			unsigned int ipm = 0;
			int j = 0;

			while ((j < (lline -1)) && (line[j] > 64))
			{
				line[j] = LPC(line[j]);
				j++;
			}

			if (sscanf( line, "whitelistconnect:%u.%u.%u.%u/%u", &ip1, &ip2, &ip3, &ip4, &ipm) == 5)
			{
				snprintf( buf, sizeof(buf), "%u.%u.%u.%u", ip1, ip2, ip3, ip4);
				entcidr = (struct CIDR *)malloc(sizeof(struct CIDR));
				entcidr->ip = inet_addr(buf);
				entcidr->masklen = ipm;
				SLIST_INSERT_HEAD(&cidrlist, entcidr, cidrs);
				syslog (LOG_DEBUG, "WhitelistConnect set %s/%d", buf, ipm);
			}
			else if (sscanf( line, "whitelistconnect:%u.%u.%u.%u", &ip1, &ip2, &ip3, &ip4) == 4)
			{
				snprintf( buf, sizeof(buf), "%u.%u.%u.%u", ip1, ip2, ip3, ip4);
				entcidr = (struct CIDR *)malloc(sizeof(struct CIDR));
				entcidr->ip = inet_addr(buf);
				entcidr->masklen = 32;
				SLIST_INSERT_HEAD(&cidrlist, entcidr, cidrs);
				syslog (LOG_DEBUG, "WhitelistConnect set %s", buf);
			}
			else if (sscanf( line, "whitelistconnect:%1024[A-Za-z0-9-.]", buf) == 1)
			{
				ent = (struct entry *)malloc(sizeof(struct entry));
				if ((ent->c = strdup(buf)) == NULL)
				{
					fclose(fh);
					UNLOCK(gwhite_lock);
					return 1;
				}
				SLIST_INSERT_HEAD(&gwhitelist, ent, entries);
				syslog (LOG_DEBUG, "WhitelistConnect set %s", buf );
			}
			else if (sscanf( line, "greylisttime:%6[0-9]", buf) == 1)
			{
				int num;

				num = atoi(buf);
				if (num > 0)
				{
					greylisttime = num;
					syslog (LOG_DEBUG, "GreylistTime set to %d seconds", greylisttime);
				}
			}
			else if (sscanf( line, "greylistexpiry:%7[0-9]", buf) == 1)
			{
				int num;

				num = atoi(buf);
				if (num > 0)
				{
					greylistexpiry = num;
					syslog (LOG_DEBUG, "GreylistExpiry set to %d seconds", greylistexpiry);
				}
			}
			else if (sscanf( line, "greylistdistance:%2[0-9]", buf) == 1)
			{
				int num;

				num = atoi(buf);
				if (num >= 0)
				{
					greylistdistance = num;
					syslog (LOG_DEBUG, "GreylistDistance set to %d", greylistdistance);
				}
			}
			else if (sscanf( line, "greylistexpiry:%7[0-9]", buf) == 1)
			{
				int num;

				num = atoi(buf);
				if (num > 0)
				{
					greylistexpiry = num;
					syslog (LOG_DEBUG, "GreylistExpiry set to %d seconds", greylistexpiry);
				}
			}
			else if (sscanf( line, "greylistasn:%6[0-9]", buf) == 1)
			{
				int num;

				num = atoi(buf);
				if (num >= 0)
				{
					ient = (struct ientry *)malloc(sizeof(struct ientry));
					ient->i = num;
					SLIST_INSERT_HEAD(&greyasnlist, ient, ientries);
					syslog (LOG_DEBUG, "GreylistASN set %d", num);
				}
			}
			else if (sscanf( line, "whitelistasn:%6[0-9]", buf) == 1)
			{
				int num;

				num = atoi(buf);
				if (num >= 0)
				{
					ient = (struct ientry *)malloc(sizeof(struct ientry));
					ient->i = num;
					SLIST_INSERT_HEAD(&whiteasnlist, ient, ientries);
					syslog (LOG_DEBUG, "WhitelistASN set %d", num);
				}
			}
			else if (sscanf( line, "whitelistconnect:/%1024[^\n]", buf) == 1)
			{
				int cflags;
				int e = 0;
				int i = 0;
				int j = lline -19;

				while (j > 0)
				{
					if (buf[j] == '/')
					{
						buf[j] = '\0';
						break;
					}

					if (buf[j] == 'e')
					{
						e = 1;
					}
					else if (buf[j] == 'i')
					{
						i = 1;
					}

					j--;
				}

				if (e == 1)
					cflags = REG_EXTENDED;
				else
					cflags = REG_BASIC;

				if (i == 1)
					cflags |= REG_ICASE;

				entre = (struct sre *)malloc(sizeof(struct sre));
				a = regcomp(&entre->re, buf, cflags);
				if ( a != 0 )
				{
					char err[2048];
					regerror(a, &entre->re, err, sizeof(err));
					fprintf(stderr, "Regex error %d in line %d %s\n", a, lineno,  err);
					fclose(fh);
					UNLOCK(gwhite_lock);
					exit(1);
				}
				SLIST_INSERT_HEAD(&whiterelist, entre, sres);
				syslog (LOG_DEBUG, "WhitelistConnect set %s", buf );

			}
			else if (sscanf( line, "greylisthost:/%1024[^\n]", buf) == 1)
			{
				int cflags;
				int e = 0;
				int i = 0;
				int j = lline -16;

				while (j > 0)
				{
					if (buf[j] == '/')
					{
						buf[j] = '\0';
						break;
					}

					if (buf[j] == 'e')
					{
						e = 1;
					}
					else if (buf[j] == 'i')
					{
						i = 1;
					}

					j--;
				}

				if (e == 1)
					cflags = REG_EXTENDED;
				else
					cflags = REG_BASIC;

				if (i == 1)
					cflags |= REG_ICASE;

				entre = (struct sre *)malloc(sizeof(struct sre));
				a = regcomp(&entre->re, buf, cflags);
				if ( a != 0 )
				{
					char err[2048];
					regerror(a, &entre->re, err, sizeof(err));
					fprintf(stderr, "Regex error %d in line %d %s\n", a, lineno,  err);
					fclose(fh);
					UNLOCK(gwhite_lock);
					exit(1);
				}
				entre->cat = 0;
				SLIST_INSERT_HEAD(&greyhostrelist, entre, sres);

				syslog (LOG_DEBUG, "Greylisthost set %s", buf );

			}
			else if (sscanf( line, "greylisthelo:/%1024[^\n]", buf) == 1)
			{
				int cflags;
				int e = 0;
				int i = 0;
				int j = lline -16;

				while (j > 0)
				{
					if (buf[j] == '/')
					{
						buf[j] = '\0';
						break;
					}

					if (buf[j] == 'e')
					{
						e = 1;
					}
					else if (buf[j] == 'i')
					{
						i = 1;
					}

					j--;
				}

				if (e == 1)
					cflags = REG_EXTENDED;
				else
					cflags = REG_BASIC;

				if (i == 1)
					cflags |= REG_ICASE;

				entre = (struct sre *)malloc(sizeof(struct sre));
				a = regcomp(&entre->re, buf, cflags);
				if ( a != 0 )
				{
					char err[2048];
					regerror(a, &entre->re, err, sizeof(err));
					fprintf(stderr, "Regex error %d in line %d %s\n", a, lineno,  err);
					fclose(fh);
					UNLOCK(gwhite_lock);
					exit(1);
				}
				entre->cat = 1;
				SLIST_INSERT_HEAD(&greyhostrelist, entre, sres);
				syslog (LOG_DEBUG, "GreylistHelo set %s", buf );
			}
			else if (sscanf( line, "greylistdnsbl:%250[A-Za-z0-9-.]", buf) == 1)
			{
				ent = (struct entry *)malloc(sizeof(struct entry));
				if ((ent->c = strdup(buf)) == NULL)
				{
					fclose(fh);
					UNLOCK(gwhite_lock);
					return 1;
				}
				SLIST_INSERT_HEAD(&dnsbllist, ent, entries);
				syslog (LOG_DEBUG, "GreylistDNSBL set %s", buf );
			}
		}
	}
	fclose(fh);
	UNLOCK(gwhite_lock);
	return 0;
}

int
ip_cidr(const unsigned long ip, const short int masklen, const unsigned long checkip)
{
	unsigned long ipaddr = 0;
	unsigned long cidrip = 0;
	unsigned long subnet = 0;

	subnet = ~0;
	subnet = subnet << (32 - masklen);

	cidrip = htonl(ip) & subnet;
	ipaddr = ntohl(checkip) & subnet;

	if (cidrip == ipaddr)
		return 1;

	return 0;
}

int
check_ip(const unsigned long checkip)
{
	int ret = 0;
	struct CIDR *entcidr;

	RDLOCK(gwhite_lock);
	SLIST_FOREACH(entcidr, &cidrlist, cidrs)
	{
		if (ip_cidr(entcidr->ip, entcidr->masklen, checkip) == 1)
		{
			syslog (LOG_INFO, "greylist skip cidr");
			ret = 1;
			break;
		}
	}
	UNLOCK(gwhite_lock);

	return ret;
}


int
check_re(const char* hostname, const char* helo)
{
	int ret = 0;
	struct sre *entre;
	struct entry *ent;
	int a;

	if ((strlen(hostname) < 1) || (strlen(helo) < 1))
		return 0;

	RDLOCK(gwhite_lock);

	SLIST_FOREACH(entre, &greyhostrelist, sres)
	{
		if (entre->cat == 0)
		{
			a = regexec(&entre->re, hostname, 0, NULL, 0);
		}
		else if (entre->cat == 1)
		{
			a = regexec(&entre->re, helo, 0, NULL, 0);
		}
		if (a == 0)
		{
			ret = 2;
			syslog (LOG_INFO, "greylist re match");
			break;
		}
		else if (a != REG_NOMATCH)
		{
				char err[2048];
				regerror(a, &entre->re, err, sizeof(err));
				syslog (LOG_ERR, "re error:  %s", err);
		}
	}

	SLIST_FOREACH(entre, &whiterelist, sres)
	{
		a = regexec(&entre->re, hostname, 0, NULL, 0);
		if (a == 0)
		{
			ret = 1;
			syslog (LOG_INFO, "greylist skip re match");
			break;
		}
		else if (a != REG_NOMATCH)
		{
				char err[2048];
				regerror(a, &entre->re, err, sizeof(err));
				syslog (LOG_ERR, "re error:  %s", err);
		}
	}
	if ((ret != 1) && (!SLIST_EMPTY(&gwhitelist)))
	{
		SLIST_FOREACH(ent, &gwhitelist, entries)
		{
			if (strcasecmp(ent->c, hostname) == 0)
			{
				syslog (LOG_INFO, "greylist skip %s", hostname);
				ret = 1;
				break;
			}
		}
	}

	UNLOCK(gwhite_lock);

	return ret;
}

int
check_asn(const int asn)
{
	int ret = 0;
	struct ientry *ient;

	RDLOCK(gwhite_lock);
	SLIST_FOREACH(ient, &greyasnlist, ientries)
	{
		if (ient->i == asn)
		{
			ret = 2;
			break;
		}
	}

	SLIST_FOREACH(ient, &whiteasnlist, ientries)
	{
		if (ient->i == asn)
		{
			ret = 1;
			break;
		}
	}
	UNLOCK(gwhite_lock);

	return ret;
}

int
unixconn(int sockfd, const char *ctlsock_path)
{
	struct sockaddr_un sa;
	int rc;
	int timeout = 1500;

	memset(&sa, 0, sizeof(sa));
	sa.sun_family = AF_UNIX;
	strlcpy(sa.sun_path, ctlsock_path, sizeof(sa.sun_path));

	rc = connect(sockfd, (struct sockaddr *)&sa, sizeof(sa));

	if(-1 == rc)
	{
		switch (errno)
		{
			case EINPROGRESS:
			case EWOULDBLOCK:
			case EINTR:
				rc = waitconnect(sockfd, timeout);
				break;

			case ECONNREFUSED:
			default:
				syslog(LOG_ERR, "no server on %s\n", ctlsock_path);
				return -1;
		}
	}

	if (0 == rc)
	{
		int len = 0;
		int slen = sizeof(len);
		if ( -1 == getsockopt(sockfd, SOL_SOCKET, SO_ERROR, (void*)&len, (void*)&slen))
		{
			int err = errno;
			if ((0 == err) || (EISCONN == err))
			{
				rc = 0;
			} else
				rc = -1;
		}
	}

	return rc;
}

int
waitconnect(int sockfd, int timeout_msec)
{
  fd_set fd;
  fd_set errfd;
  struct timeval interval;
  int rc;

  /* now select() until we get connect or timeout */
  FD_ZERO(&fd);
  FD_SET(sockfd, &fd);

  FD_ZERO(&errfd);
  FD_SET(sockfd, &errfd);

  interval.tv_sec = timeout_msec/1000;
  timeout_msec -= interval.tv_sec*1000;

  interval.tv_usec = timeout_msec*1000;

  rc = select(sockfd+1, NULL, &fd, &errfd, &interval);
  if(-1 == rc)
    /* error, no connect here, try next */
    return -1;

  else if(0 == rc)
    /* timeout, no connect today */
    return 1;

  if(FD_ISSET(sockfd, &errfd)) {
    /* error condition caught */
    return 2;
  }

  /* we have a connect! */
  return 0;
}

int
dothost(char *host)
{
	int j = strlen(host);
	int i = j;
	int dot = 0;
	int hyphen = 0;

	if (j < 5)
		return 0;

	while (i > 0)
	{
		if (host[i] == '.')
		{
			dot++;
			if (dot > 2)
			{
				return (dot);
			}
		}
		else if (host[i] == '-')
		{
			hyphen++;
			if (hyphen > 1)
			{
				return( dot + hyphen);
			}
		}
		i--;
	}

	return (dot);
}

int qrydnsasn( const unsigned int ip) {
	unsigned char dnsreply[PACKETSZ];
	unsigned char expdn[PACKETSZ];
	unsigned char origin[6];
	unsigned char *p;
	unsigned char host[128];
	HEADER *dnsheader;
	int ret = 0;
	int dxp, i;
	unsigned int ttl;

	memset( dnsreply, '\0', sizeof( dnsreply ));
	WRLOCK(gwhite_lock);
	if (res_init() == -1)
	{
		UNLOCK(gwhite_lock);
		return -1;
	}

#ifdef ASN32BIT
	snprintf((char *)host, 127, "%u.%u.%u.%u.%s",
			(ip >> 24) & 0xff,
			(ip >> 16) & 0xff,
			(ip >> 8) & 0xff,
			(ip) & 0xff,
			 "origin.asn.cymru.com");
#else
	snprintf((char *)host, 127, "%u.%u.%u.%s",
			(ip >> 16) & 0xff,
			(ip >> 8) & 0xff,
			(ip) & 0xff,
			"origin.asn.cymru.com");
#endif /* ASN32BIT */

	ret = res_query( (char *) host, C_IN, T_TXT, dnsreply, sizeof( dnsreply ));
	UNLOCK(gwhite_lock);
	if (ret == -1)
		return -2;

	dnsheader = (HEADER *) &dnsreply;
	if (dnsheader->rcode == 0) {
		p = dnsreply +sizeof(HEADER);

		if((dxp = dn_expand(dnsreply, dnsreply + ret, p, expdn, PACKETSZ)) < 0) {
			return -3;
		}

		p += dxp;

		GETSHORT(i, p);
		if (i != T_TXT)
			return -1;

		p += INT16SZ;

		if((dxp = dn_expand(dnsreply, dnsreply + ret, p, expdn, PACKETSZ)) < 0) {
			return -4;
		}

		p += dxp;

		GETSHORT(i, p);
		if (i != T_TXT)
			return -5;

		p += INT16SZ;

		GETLONG(ttl, p);
		GETSHORT(i, p);

		if (i < 10)
			return -6;

		p++;
		strncpy((char *)origin, (const char *)p, 5);

		ret = atoi((const char *)origin);
	}

	return ret;
}

int qrydnsa( char *host) {
	unsigned char dnsreply[PACKETSZ];
	unsigned char expdn[PACKETSZ];
	unsigned char *p;
	HEADER *dnsheader;
	int dxp, i;
	int ret = -1;

	memset( dnsreply, '\0', sizeof( dnsreply ));

	WRLOCK(gwhite_lock);
	if (res_init() != -1)
	{
		ret = res_query( host, C_IN, T_A, dnsreply, sizeof( dnsreply ));
		UNLOCK(gwhite_lock);

		if (ret == -1)
			return -2;

		dnsheader = (HEADER *) &dnsreply;
		if (dnsheader->rcode == 0)
		{
			p = dnsreply +sizeof(HEADER);

			if((dxp = dn_expand(dnsreply, dnsreply + ret, p, expdn, PACKETSZ)) < 0) {
				return -3;
			}

			p += dxp;

			GETSHORT(i, p);
			if (i != T_A)
				return -1;

			p += INT16SZ;

			if((dxp = dn_expand(dnsreply, dnsreply + ret, p, expdn, PACKETSZ)) < 0) {
				return -4;
			}

			p += dxp;

			GETSHORT(i, p);
			if (i == T_A)
				return 1;
		}

	} else
		UNLOCK(gwhite_lock);

	return ret;
}

int qrydnsbl( const unsigned int ip) {

	int ret = 0;
	struct entry *ent;
	char hostn[128];

	SLIST_FOREACH(ent, &dnsbllist, entries)
	{
		snprintf((char *)hostn, 127, "%u.%u.%u.%u.%s",
			(ip >> 24) & 0xff,
			(ip >> 16) & 0xff,
			(ip >> 8) & 0xff,
			(ip) & 0xff,
			 ent->c);

		if (qrydnsa(hostn) == 1)
		{
			syslog (LOG_INFO, "greylist dnsbl match %s", ent->c);
			ret = 1;
			break;
		}
	}

	return ret;
}

#ifndef NOMXEVAL
unsigned int qrydnsmx( char *host) {
	unsigned char dnsreply[PACKETSZ];
	unsigned char expdn[PACKETSZ];
	unsigned char *p;
	HEADER *dnsheader;
	int ret = 0;
	int dxp, i;
	unsigned int ttl;
	struct rrmx mx;

	memset( dnsreply, '\0', sizeof( dnsreply ));
	mx.hostname = NULL;

	WRLOCK(gwhite_lock);
	if (res_init() == -1)
	{
		UNLOCK(gwhite_lock);
		return 1;
	}

	ret = res_query( host, C_IN, T_MX, dnsreply, sizeof( dnsreply ));
	UNLOCK(gwhite_lock);
	if (ret == -1)
		return 2;

	dnsheader = (HEADER *) &dnsreply;
	if (dnsheader->rcode == 0) {
		p = dnsreply +sizeof(HEADER);

		if((dxp = dn_expand(dnsreply, dnsreply + ret, p, expdn, PACKETSZ)) < 0) {
			return 3;
		}

		p += dxp;

		GETSHORT(i, p);
		if (i != T_MX)
			return 1;

		p += INT16SZ;

		while (p < dnsreply + ret) {

			if((dxp = dn_expand(dnsreply, dnsreply + ret, p, expdn, PACKETSZ)) < 0) {
				return 4;
			}

			p += dxp;

			GETSHORT(i, p);
			if (i != T_MX)
				break;

			p += INT16SZ;

			GETLONG(ttl, p);
			GETSHORT(i, p);


			if((dxp = dn_expand(dnsreply, dnsreply + ret, p + 2, expdn, PACKETSZ)) > 0)
			{
				if (mx.hostname != NULL)
				{
					if (mx.pref > ((p[0] << 8) | p[1]))
					{
						free(mx.hostname);
						mx.hostname = strdup(expdn);
						mx.pref = (p[0] << 8) | p[1];
					}
				} else {
					mx.hostname = strdup(expdn);
					mx.pref = (p[0] << 8) | p[1];
				}

			}

			p +=  i;
		}
	}

	if ((mx.hostname == NULL) || (strlen(mx.hostname) < 4))
	{
		if (mx.hostname != NULL)
			free(mx.hostname);

		return 1;
	}

	memset( dnsreply, '\0', sizeof( dnsreply ));

	WRLOCK(gwhite_lock);
	ret = res_query( mx.hostname, C_IN, T_A, dnsreply, sizeof( dnsreply ));
	UNLOCK(gwhite_lock);
	free(mx.hostname);
	if (ret == -1)
		return 2;

	dnsheader = (HEADER *) &dnsreply;
	if (dnsheader->rcode == 0) {
		p = dnsreply +sizeof(HEADER);

		if((dxp = dn_expand(dnsreply, dnsreply + ret, p, expdn, PACKETSZ)) < 0) {
			return 3;
		}

		p += dxp;

		GETSHORT(i, p);
		if (i != T_A)
			return 1;

		p += INT16SZ;


		if((dxp = dn_expand(dnsreply, dnsreply + ret, p, expdn, PACKETSZ)) < 0) {
			return 4;
		}

		p += dxp;

		GETSHORT(i, p);
		if (i != T_A)
			return 0;

		p += INT16SZ;

		GETLONG(ttl, p);
		GETSHORT(i, p);

		if (i == 4)
		{
			struct in_addr inaddr;

			memcpy( (char *)&inaddr, p, INADDRSZ);
			return inaddr.s_addr;
		}
	}

	return 0;
}

char* rfc822domain( char* rfc822addr )
{
	char *p = rfc822addr;
	char *q;
	int k;

	k = strlen( p );

	if ( k >= 1 )
	{
		for ( q = p + k - 1; q != p; --q )
		{
			if ( *(q - 1) == '@' )
			{
				p = q;
				break;
			}
		}
	}

	return p;
}

int mxeval(char *mailaddr, const unsigned int ip)
{
	char * domain;
	unsigned int mxa;

	domain = rfc822domain(mailaddr);
	if ((domain != NULL) && (strlen(domain) > 3))
	{
		mxa = qrydnsmx(domain);
		if ((mxa > 10) && (mxa == ip))
			return 1;
	}
	return 0;
}
#endif

#ifdef POPAUTH

int
checkpopauth(const char *dbfile, struct sockaddr *ip)
{
	int ret = 0;
	DB *db;
	struct sockaddr_in *sin;
	struct in_addr addr;
	DBT d;
	DBT q;
	char ipbuf[17];

	/* skip anything not IPv4 (for now) */
	if (ip->sa_family != AF_INET)
		return 0;
	else
		sin = (struct sockaddr_in *)ip;

	memset(&d, 0, sizeof d);
	memset(&q, 0, sizeof q);

	memcpy(&addr.s_addr, &sin->sin_addr, sizeof addr.s_addr);

	inet_ntop(AF_INET, &addr, ipbuf, sizeof ipbuf);

#ifdef POPTEXT
	q.data = sprintf("POP:%s",ipbuf);
#else
	q.data = ipbuf;
#endif
	q.size = strlen(q.data);
#if DB_VERSION_MAJOR > 3
	if (db_create(&db, NULL, 0) != 0 ||
	    db->open(db, NULL, dbfile, NULL, DB_HASH, DB_RDONLY, 0) != 0)
#elif DB_VERSION_MAJOR == 3
	if (db_create(&db, NULL, 0) != 0 ||
	    db->open(db, dbfile, NULL, DB_HASH, DB_RDONLY, 0) != 0)
#elif DB_VERSION_MAJOR == 2
	if (db_open(dbfile, DB_HASH, DB_RDONLY, DB_MODE, NULL, NULL, &db) != 0)
#else /* DB_VERSION_MAJOR == 2 */
	if ((db = db_open(dbfile, O_RDONLY, DB_MODE, DB_HASH, NULL)) == NULL)
#endif /* DB_VERSION_MAJOR < 2 */
		return 0;

#if DB_VERSION_MAJOR < 2
	if (!((db->get) (db, &q, &d, 0)))
#else /* DB_VERSION_MAJOR < 2 */
	if (!((db->get) (db, NULL, &q, &d, 0)))
#endif
		ret = 1;

#if DB_VERSION_MAJOR < 2
	(void) db->close(db);
#else /* DB_VERSION_MAJOR < 2 */
	(void) db->close(db, 0);
#endif

	return ret;
}
#endif

